
from .specification import Specification

class SectionSpecification(Specification):
    """
    TBC
    """


    def __init__(self, name, extras):

        super(SectionSpecification, self).__init__(name, None, None, extras)


    def __str__(self):

        return "<%s.%s: name=%s; help=%s; aliases=%s; extras=%s>" %\
            (self.__module__, self.__class__.__name__, self.name, self.help, self.aliases, self.extras)


def section(name, **kwargs):
    """
    Creates a section specification from the given parameters
    """

    extras  =   None

    for n, v in kwargs.items():

        if False:

            pass
        elif 'extras' == n:

            extras = v
        else:

            raise TypeError("`section` method does not recognise the `%s` keyword argument" % (n, ))

    return SectionSpecification(name, extras)






from .arguments import Arguments
from .util import _dict_get_N, _get_program_name

from .specification import Specification
from .flag_specification import FlagSpecification
from .option_specification import OptionSpecification
from .section_specification import SectionSpecification

import os
import re
import sys


def _ensure_single_space_prefix(s):

    if s:

        s = str(s).strip()

        if 0 != len(s):

            s = ' ' + s

    return s


def _generate_version_string(argv, options):

    program_name    =   _get_program_name(argv, options)
    version_prefix  =   _dict_get_N((options, ''), 'version_prefix', 'version-prefix')

    version         =   options.get('version')

    if version:

        if isinstance(version, (list, tuple, )):

            version =   [ str(i) for i in version ]

            version =   '.'.join(version)
    else:

        version_major   =  _dict_get_N(options, 'version_major', 'version-major')

        if not version_major:

            raise ValueError("options must specify 'version' or 'version_major' [ + 'version_minor' [ + [ 'version_patch' (or 'version_revision') [ + 'version_build' ] ] ] ")

        version_minor   =   _dict_get_N(options, 'version_minor', 'version-minor')
        version_patch   =   _dict_get_N(options, 'version_patch', 'version-patch', 'version_revision', 'version-revision')
        version_build   =   _dict_get_N(options, 'version_build', 'version-build')

        version         =   str(version_major)

        if version_minor:

            version     +=  '.' + str(version_minor)

        if version_patch:

            version     +=  '.' + str(version_patch)

        if version_build:

            version     +=  '.' + str(version_build)

    return "%s %s%s" % (program_name, version_prefix, version)


def show_usage(specifications, **kwargs):
    """
    Displays program usage from the given `specifications` (or arguments), according to the given options
    """

    argv                =   sys.argv

    if isinstance(specifications, (Arguments, )):

        args            =   specifications
        argv            =   args.argv
        specifications  =   args.specifications

    if __debug__:

        if specifications == None:

            raise TypeError("`specifications` may not be `None`")
        elif isinstance(specifications, (list, tuple, )):

            for index, a in enumerate(specifications):

                if not isinstance(a, (Specification, )):

                    raise TypeError("every element in `specifications` must be an instance of `clasp.Specification`: the element at index %d is of type `%s`" % (index, type(a).__name__))
        else:

            raise TypeError("`specifications` must be a `list` or a `tuple`")

        alias_dups = {}
        for spec in specifications:

            for a in spec.aliases:

                if a in alias_dups:

                    sys.stderr.write("WARNING: alias '%s' is already used for specification '%s'\n" % (a, spec))
                else:

                    alias_dups[a] = spec

    options             =   kwargs

    exit_code           =   _dict_get_N(options, 'exit', 'exit_code', 'exit-code')
    flags_and_options   =   _dict_get_N((options, ' [ ... flags and options ... ]'), 'flags_and_options', 'flags-and-options')
    info_lines          =   _dict_get_N(options, 'info_lines', 'info-lines')
    program_name        =   _get_program_name(argv, options)
    stream              =   _dict_get_N((options, sys.stdout), 'stream')
    suppress_blanks     =   _dict_get_N(options, 'suppress_blank_lines_between_options', 'suppress-blank-lines-between-options')
    values              =   _dict_get_N((options, ''), 'values_string', 'values-string', 'values', default_if_none=True)

    if not info_lines:

        info_lines  =   []
    elif isinstance(info_lines, (list, tuple, )):

        pass
    else:

        info_lines  =   [ info_lines ]

    info_lines      =   [ _generate_version_string(argv, options) if l in ( ':version', ':version:' ) else l for l in info_lines ]

    flags_and_options   =   _ensure_single_space_prefix(flags_and_options)
    values              =   _ensure_single_space_prefix(values)

    # sift the specifications to sort out which are value-option aliases (VOAs)

    pure_specifications    =   []
    voas            =   {}

    for specification in specifications:

        m       =   re.match(r'(-+[a-zA-Z0-3_-]+)[=:](.+)$', specification.name)

        if m:

            name    =   m.group(1)
            value   =   m.group(2)

            if not voas.get(name):

                voas[name] = []

            voas[name].append([ specification, value ])
        else:

            pure_specifications.append(specification)

    specifications         =   pure_specifications


    for info_line in info_lines:

        stream.write("%s\n" % info_line)

    stream.write("USAGE: %s%s%s\n\n" % (program_name, flags_and_options, values))

    if specifications:

        stream.write("flags/options:\n\n")

        for specification in specifications:

            if isinstance(specification, FlagSpecification):

                for a2 in specification.aliases:

                    stream.write("\t%s\n" % a2)

                stream.write("\t%s\n" % specification.name)
                stream.write("\t\t%s\n" % specification.help)
            elif isinstance(specification, OptionSpecification):

                voa = voas.get(specification.name)

                if voa:

                    for a2 in voa:

                        for a3 in a2[0].aliases:

                            stream.write("\t%s %s\n" % (a3, a2[0].name))

                for a2 in specification.aliases:

                    stream.write("\t%s <value>\n" % a2)

                stream.write("\t%s=<value>\n" % specification.name)
                stream.write("\t\t%s\n" % specification.help)
                if specification.values_range:

                    stream.write("\t\twhere <value> one of:\n")

                    for v in specification.values_range:

                        stream.write("\t\t\t%s\n" % v)

            elif isinstance(specification, SectionSpecification):

                stream.write("\t%s\n" % specification.name)

            elif isinstance(specification, Specification):

                pass
            else:

                pass

            if not suppress_blanks:

                stream.write("\n")

    if None != exit_code:

        sys.exit(exit_code)


def show_version(specifications, **kwargs):
    """
    Displays program version from the given `specifications` (or arguments), according to the given options
    """

    argv                =   sys.argv

    if isinstance(specifications, (Arguments, )):

        args            =   specifications
        argv            =   args.argv
        specifications  =   args.specifications

    if __debug__:

        if specifications == None:

            raise TypeError("`specifications` may not be `None`")
        elif isinstance(specifications, (list, tuple, )):

            for index, a in enumerate(specifications):

                if not isinstance(a, (Specification, )):

                    raise TypeError("every element in `specifications` must be an instance of `clasp.Specification`: the element at index %d is of type `%s`" % (index, type(a).__name__))
        else:

            raise TypeError("`specifications` must be a `list` or a `tuple`")

    # options:
    #
    # - exit (integer, or None)
    # - program_name (string)
    # - stream (file, ...)
    # - version (array) or
    # - version_major, version_minor, version_patch or version_revision, version_build (all integer)
    # - version_prefix (string)

    exit_code       =   kwargs.get('exit', kwargs.get('exit_code', kwargs.get('exit-code')))
    stream          =   kwargs.get('stream', sys.stdout)

    version         =   _generate_version_string(argv, kwargs)

    stream.write("%s\n" % version)

    if None != exit_code:

        sys.exit(exit_code)


# ############################## end of file ############################# #






from .specification import Specification
from .util import _MULTIPLE_OPTION_ACTION_DEFAULT, _MULTIPLE_ACTION_OPTION_ALLOWED

from .util import _SUPPORT_long

class OptionSpecification(Specification):
    """
    TBC
    """


    if _SUPPORT_long:

        _VALID_VALUE_TYPES = (bool, float, int, long, str, )
    else:

        _VALID_VALUE_TYPES = (bool, float, int, str, )


    def __init__(self, name, aliases, help, extras, values_range, default_value, is_required, require_message, value_type, on_multiple):

        super(OptionSpecification, self).__init__(name, aliases, help, extras)

        self.values_range       =   values_range
        self.default_value      =   default_value
        self.is_required        =   is_required
        self.require_message    =   require_message
        self.value_type         =   value_type
        self.on_multiple        =   on_multiple


    def __str__(self):

        return "<%s.%s: name=%s; help=%s; aliases=%s; extras=%s, default_value=%s, value_type=%s, values_range=%s, on_multiple=%s, is_required=%s, require_message=%s >" %\
            (self.__module__, self.__class__.__name__, self.name, self.help, self.aliases, self.extras, self.default_value, self.value_type, self.values_range, self.on_multiple, self.is_required, self.require_message,)


def option(name, **kwargs):
    """
    Creates an option specification from the given parameters
    """

    aliases =   None
    help    =   None
    extras  =   None

    values_range    =   None
    default_value   =   None
    is_required     =   None
    require_message =   None
    value_type      =   None
    on_multiple     =   _MULTIPLE_OPTION_ACTION_DEFAULT

    for n, v in kwargs.items():

        if False:

            pass
        elif 'alias' == n:

            aliases = ( v, )
        elif 'aliases' == n:

            aliases = v
        elif 'help' == n:

            help = v
        elif 'extras' == n:

            extras = v
        elif 'values_range' == n or 'values' == n:

            values_range = v
        elif 'default_value' == n or 'default' == n:

            default_value = v
        elif 'required' == n or 'is_required' == n:

            is_required = v
        elif 'on_multiple' == n:

            if v is None:

                pass
            elif str == type(v):

                v = v.upper()

                if v not in _MULTIPLE_ACTION_OPTION_ALLOWED:

                    raise TypeError("`option` method keyword argument `on_multiple` must be one of %s (in any case); '%s' given" % (_MULTIPLE_ACTION_OPTION_ALLOWED, v))

            else:

                raise TypeError("`option` method keyword argument `on_multiple` must be `None` or a string; '%s' (%s) given" % (v, type(v)))

            on_multiple = v
        elif 'require_message' == n:

            require_message = v
        elif 'value_type' == n:

            value_type = v

            if value_type is None:

                pass
            elif value_type in OptionSpecification._VALID_VALUE_TYPES:

                pass
            else:

                raise TypeError("`option` method supports `value_type` only for `None` and the types %s; `%s` (%s) given" % ([t.__name__ for t in OptionSpecification._VALID_VALUE_TYPES], value_type, type(value_type)))
        else:

            raise TypeError("`option` method does not recognise the `%s` keyword argument" % (n, ))

    return OptionSpecification(name, aliases, help, extras, values_range, default_value, is_required, require_message, value_type, on_multiple)





import os

import sys


_SUPPORT_long = True
if sys.version_info >= (3, 0):

    _SUPPORT_long = False


_MULTIPLE_ACTION_ALLOW      =   'ALLOW'
_MULTIPLE_ACTION_IGNORE     =   'IGNORE'
_MULTIPLE_ACTION_REJECT     =   'REJECT'
_MULTIPLE_ACTION_REPLACE    =   'REPLACE'

_MULTIPLE_ACTION_FLAG_ALLOWED = (

    _MULTIPLE_ACTION_IGNORE,
    _MULTIPLE_ACTION_REJECT,
    _MULTIPLE_ACTION_REPLACE,
)

_MULTIPLE_ACTION_OPTION_ALLOWED = (

    _MULTIPLE_ACTION_ALLOW,
    _MULTIPLE_ACTION_IGNORE,
    _MULTIPLE_ACTION_REJECT,
    _MULTIPLE_ACTION_REPLACE,
)

_MULTIPLE_FLAG_ACTION_DEFAULT    =   _MULTIPLE_ACTION_REPLACE
_MULTIPLE_OPTION_ACTION_DEFAULT  =   _MULTIPLE_ACTION_ALLOW


def _dict_get_N(d, *keys, **kwargs):

    default         =   None
    default_if_none =   kwargs.get('default_if_none', False)


    if isinstance(d, (tuple, )):

        default =   d[1]
        d       =   d[0]

    for key in keys:

        if key in d:

            v = d.get(key)

            if None == v and default_if_none:

                continue

            return v

    return default


def _get_program_name(argv, options):

    program_name    =   _dict_get_N(options, 'program_name', 'program-name')

    if not program_name:

        bn              =   os.path.basename(argv[0])

        program_name    =   bn

    return program_name


def _global_multiple_flags_action():

    a = os.environ.get('CLASP_MULTIPLE_FLAG_ACTION')

    if a:

        a = a.upper()

    if a:

        if not a in _MULTIPLE_ACTION_FLAG_ALLOWED:

            a = None

    if not a:

        a = _MULTIPLE_FLAG_ACTION_DEFAULT

    return a


def _global_multiple_options_action():

    a = os.environ.get('CLASP_MULTIPLE_OPTION_ACTION')

    if a:

        a = a.upper()

    if a:

        if not a in _MULTIPLE_ACTION_OPTION_ALLOWED:

            a = None

    if not a:

        a = _MULTIPLE_OPTION_ACTION_DEFAULT

    return a





#!/usr/bin/env python

# examples/show_usage_and_version.py


# imports

import pyclasp as clasp

import sys

# constants

VERSION = [ 0, 0, 2 ]

INFO_LINES = (

    'CLASP.Python examples',
    ':version:',
    "Illustrates use of CLASP.Python's clasp.show_usage() and clasp.show_version() methods",
    '',
)

# Specify specifications, parse, and checking standard flags

specifications = (

    clasp.HelpFlag(),
    clasp.VersionFlag(),
)

args = clasp.parse(sys.argv, specifications)

if args.flag_is_specified('--help'):

    clasp.show_usage(specifications, exit_code=0, version=VERSION, stream=sys.stdout, info_lines=INFO_LINES)

if args.flag_is_specified('--version'):

    clasp.show_version(specifications, exit_code=0, version=VERSION, stream=sys.stdout)


# Check for any unrecognised flags or options

unused = args.get_first_unused_flag_or_option();
if (unused):

    sys.stderr.write("%s: unrecognised flag/option: %s\n" % (args.program_name, unused))

    sys.exit(1)


sys.stdout.write("no flags specified\n");




#! /usr/bin/env python

from pyclasp import Arguments
from pyclasp import specification, option
from pyclasp import Flag
from pyclasp import Option
import pyclasp as clasp

import unittest

class Options_tester_1(unittest.TestCase):

    def test_option_with_given_value_in_same_argument(self):

        values_range    =   (

            'silent',
            'normal',
            'verbose',
        )
        option_verbose    =   clasp.option('--verbose', alias = '-v', values_range=values_range, default_value='normal')

        specifications =   (

            option_verbose,
        )
        argv    =   ( 'myprog', '--verbose=silent' )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertFalse(args.flags)
        self.assertTrue(args.options)
        self.assertFalse(args.values)

        option  =   args.options[0]

        self.assertEqual('silent', option.value)

    def _test_option_with_given_value_in_separate_argument(self):

        values_range    =   (

            'silent',
            'normal',
            'verbose',
        )
        option_verbose    =   clasp.option('--verbose', alias = '-v', values_range=values_range, default_value='normal')

        specifications =   (

            option_verbose,
        )
        argv    =   ( 'myprog', '--verbose', 'silent' )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertFalse(args.flags)
        self.assertTrue(args.options)
        self.assertFalse(args.values)

        option  =   args.options[0]

        self.assertEqual('silent', option.value)

    def test_option_with_empty_value(self):

        values_range    =   (

            'silent',
            'normal',
            'verbose',
        )
        option_verbose    =   clasp.option('--verbose', alias = '-v', values_range=values_range, default_value='normal')

        specifications =   (

            option_verbose,
        )
        argv    =   ( 'myprog', '--verbose=' )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertFalse(args.flags)
        self.assertTrue(args.options)
        self.assertFalse(args.values)

        option  =   args.options[0]

        self.assertEqual('normal', option.value)

    def test_option_with_no_given_value(self):

        values_range    =   (

            'silent',
            'normal',
            'verbose',
        )
        option_verbose    =   clasp.option('--verbose', alias = '-v', values_range=values_range, default_value='normal')

        specifications =   (

            option_verbose,
        )
        argv    =   ( 'myprog', '--verbose' )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertFalse(args.flags)
        self.assertTrue(args.options)
        self.assertFalse(args.values)

        option  =   args.options[0]

        self.assertEqual('normal', option.value)







class Specification(object):
    """
    TBC
    """


    def __init__(self, name, aliases, help, extras):

        if not isinstance(name, str):

            raise TypeError("`name` must be of type `str`")


        if aliases:

            if not isinstance(aliases, ( list, tuple )):

                raise TypeError("`aliases` must be `None` or an instance of `list` or `tuple`")
        else:

            aliases = ()


        if help:

            if not isinstance(help, ( str, )):

                raise TypeError("`help` must be an instance of `str`")


        if extras:

            if not isinstance(extras, ( dict, )):

                raise TypeError("`extras` must be `None` or an instance of `dict`")

        else:

            extras  =   {}


        self.name       =   name
        self.aliases    =   aliases
        self.help       =   help
        self.extras     =   extras


def specification(name, **kwargs):
    """
    TBC
    """


    aliases =   None
    help    =   None
    extras  =   None

    for n, v in kwargs.items():

        if 'alias' == n:

            aliases = ( v, )
        elif 'aliases' == n:

            aliases = v
        elif 'help' == n:

            help = v
        elif 'extras' == n:

            extras = v
        else:

            raise TypeError("`specification` method does not recognise the `%s` keyword argument" % (n, ))

    return Specification(name, aliases, help, extras)





#!/usr/bin/env python

from pyclasp import Arguments
from pyclasp import specification, option
from pyclasp import Flag
from pyclasp import Option
import pyclasp as clasp

import unittest

class Arguments_tester_1(unittest.TestCase):


    def test_empty_args_via_clasp_parse(self):

        argv    =   ()

        with self.assertRaises(IndexError):

            clasp.parse(argv)


    def test_no_args_via_clasp_parse(self):

        argv    =   ( 'myprog', )
        args    =   clasp.parse(argv)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple ))
        self.assertFalse(args.flags)

        self.assertIsInstance(args.options, ( tuple ))
        self.assertFalse(args.options)

        self.assertIsInstance(args.values, ( tuple ))
        self.assertFalse(args.values)


    def test_no_args_via_Arguments_constructor(self):

        argv    =   ( 'myprog', )
        args    =   Arguments(argv)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple ))
        self.assertFalse(args.flags)

        self.assertIsInstance(args.options, ( tuple ))
        self.assertFalse(args.options)

        self.assertIsInstance(args.values, ( tuple ))
        self.assertFalse(args.values)


    def test_one_value(self):

        argv    =   ( 'myprog', 'value1', )
        args    =   clasp.parse(argv)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple ))
        self.assertFalse(args.flags)

        self.assertIsInstance(args.options, ( tuple ))
        self.assertFalse(args.options)

        self.assertIsInstance(args.values, ( tuple ))
        self.assertTrue(args.values)
        self.assertEqual(1, len(args.values))


    def test_two_values(self):

        argv    =   ( 'myprog', 'value1', 'value2' )
        args    =   clasp.parse(argv)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple ))
        self.assertFalse(args.flags)

        self.assertIsInstance(args.options, ( tuple ))
        self.assertFalse(args.options)

        self.assertIsInstance(args.values, ( tuple ))
        self.assertTrue(args.values)


    def test_ten_values(self):

        argv    =   [ 'myprog', ] + [ "value%d" % i for i in range(0, 10) ]
        args    =   clasp.parse(argv)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple ))
        self.assertFalse(args.flags)

        self.assertIsInstance(args.options, ( tuple ))
        self.assertFalse(args.options)

        self.assertIsInstance(args.values, ( tuple ))
        self.assertTrue(args.values)
        self.assertEqual(10, len(args.values))


    def test_one_flag(self):

        argv    =   ( 'myprog', '-f1', )
        args    =   clasp.parse(argv)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple ))
        self.assertTrue(args.flags)
        self.assertEqual(1, len(args.flags))

        flag    =   args.flags[0]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-f1')
        self.assertIsNone(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'f1')
        self.assertEqual(flag.name              ,   '-f1')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '-f1')
        self.assertEqual(flag                   ,   '-f1')

        self.assertIsInstance(args.options, ( tuple ))
        self.assertFalse(args.options)

        self.assertIsInstance(args.values, ( tuple ))


    def test_two_flags(self):

        argv    =   ( 'myprog', '-f1', '--flag2' )
        args    =   clasp.parse(argv)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple ))
        self.assertTrue(args.flags)
        self.assertEqual(2, len(args.flags))

        flag    =   args.flags[0]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-f1')
        self.assertIsNone(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'f1')
        self.assertEqual(flag.name              ,   '-f1')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '-f1')
        self.assertEqual(flag                   ,   '-f1')

        flag    =   args.flags[1]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   2)
        self.assertEqual(flag.given_name        ,   '--flag2')
        self.assertIsNone(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   2)
        self.assertEqual(flag.given_label       ,   'flag2')
        self.assertEqual(flag.name              ,   '--flag2')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '--flag2')
        self.assertEqual(flag                   ,   '--flag2')

        self.assertIsInstance(args.options, ( tuple ))
        self.assertFalse(args.options)

        self.assertIsInstance(args.values, ( tuple ))
        self.assertFalse(args.values)


    def test_three_flags(self):

        argv    =   ( 'myprog', '-f1', '--flag2', '---x' )
        args    =   clasp.parse(argv)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple ))
        self.assertTrue(args.flags)
        self.assertEqual(3, len(args.flags))

        flag    =   args.flags[0]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-f1')
        self.assertIsNone(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'f1')
        self.assertEqual(flag.name              ,   '-f1')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '-f1')
        self.assertEqual(flag                   ,   '-f1')

        flag    =   args.flags[1]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   2)
        self.assertEqual(flag.given_name        ,   '--flag2')
        self.assertIsNone(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   2)
        self.assertEqual(flag.given_label       ,   'flag2')
        self.assertEqual(flag.name              ,   '--flag2')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '--flag2')
        self.assertEqual(flag                   ,   '--flag2')

        flag    =   args.flags[2]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   3)
        self.assertEqual(flag.given_name        ,   '---x')
        self.assertIsNone(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   3)
        self.assertEqual(flag.given_label       ,   'x')
        self.assertEqual(flag.name              ,   '---x')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '---x')
        self.assertEqual(flag                   ,   '---x')

        self.assertIsInstance(args.options, ( tuple ))
        self.assertFalse(args.options)

        self.assertIsInstance(args.values, ( tuple ))
        self.assertFalse(args.values)


    def test_one_option(self):

        argv    =   ( 'myprog', '-o1=v1', )
        args    =   clasp.parse(argv)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple ))
        self.assertFalse(args.flags)

        self.assertIsInstance(args.options, ( tuple ))
        self.assertTrue(args.options)
        self.assertEqual(1, len(args.options))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   1)
        self.assertEqual(option.given_name      ,   '-o1')
        self.assertIsNone(option.argument_specification)
        self.assertEqual(option.given_hyphens   ,   1)
        self.assertEqual(option.given_label     ,   'o1')
        self.assertEqual(option.name            ,   '-o1')
        self.assertEqual(option.value           ,   'v1')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '-o1=v1')
        self.assertEqual(option                 ,   '-o1=v1')

        self.assertIsInstance(args.values, ( tuple ))
        self.assertFalse(args.values)


    def test_two_options(self):

        argv    =   ( 'myprog', '-o1=v1', '--option2=value2' )
        args    =   clasp.parse(argv)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple ))
        self.assertFalse(args.flags)

        self.assertIsInstance(args.options, ( tuple ))
        self.assertTrue(args.options)
        self.assertEqual(2, len(args.options))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   1)
        self.assertEqual(option.given_name      ,   '-o1')
        self.assertIsNone(option.argument_specification)
        self.assertEqual(option.given_hyphens   ,   1)
        self.assertEqual(option.given_label     ,   'o1')
        self.assertEqual(option.name            ,   '-o1')
        self.assertEqual(option.value           ,   'v1')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '-o1=v1')
        self.assertEqual(option                 ,   '-o1=v1')

        option  =   args.options[1]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   2)
        self.assertEqual(option.given_name      ,   '--option2')
        self.assertIsNone(option.argument_specification)
        self.assertEqual(option.given_hyphens   ,   2)
        self.assertEqual(option.given_label     ,   'option2')
        self.assertEqual(option.name            ,   '--option2')
        self.assertEqual(option.value           ,   'value2')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--option2=value2')
        self.assertEqual(option                 ,   '--option2=value2')

        self.assertIsInstance(args.values, ( tuple ))
        self.assertFalse(args.values)


    def test_three_options(self):

        argv    =   ( 'myprog', '-o1=v1', '--option2=value2', '---the-third-option=the third value' )
        args    =   clasp.parse(argv)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple ))
        self.assertFalse(args.flags)

        self.assertIsInstance(args.options, ( tuple ))
        self.assertTrue(args.options)
        self.assertEqual(3, len(args.options))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   1)
        self.assertEqual(option.given_name      ,   '-o1')
        self.assertIsNone(option.argument_specification)
        self.assertEqual(option.given_hyphens   ,   1)
        self.assertEqual(option.given_label     ,   'o1')
        self.assertEqual(option.name            ,   '-o1')
        self.assertEqual(option.value           ,   'v1')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '-o1=v1')
        self.assertEqual(option                 ,   '-o1=v1')

        option  =   args.options[1]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   2)
        self.assertEqual(option.given_name      ,   '--option2')
        self.assertIsNone(option.argument_specification)
        self.assertEqual(option.given_hyphens   ,   2)
        self.assertEqual(option.given_label     ,   'option2')
        self.assertEqual(option.name            ,   '--option2')
        self.assertEqual(option.value           ,   'value2')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--option2=value2')
        self.assertEqual(option                 ,   '--option2=value2')

        option  =   args.options[2]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   3)
        self.assertEqual(option.given_name      ,   '---the-third-option')
        self.assertIsNone(option.argument_specification)
        self.assertEqual(option.given_hyphens   ,   3)
        self.assertEqual(option.given_label     ,   'the-third-option')
        self.assertEqual(option.name            ,   '---the-third-option')
        self.assertEqual(option.value           ,   'the third value')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '---the-third-option=the third value')
        self.assertEqual(option                 ,   '---the-third-option=the third value')

        self.assertIsInstance(args.values, ( tuple ))
        self.assertFalse(args.values)


    def test_one_flag_and_one_option_and_one_value(self):

        argv    =   ( 'myprog', '-f1', 'value1', '--first-option=val1' )
        args    =   clasp.parse(argv)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertTrue(args.flags)
        self.assertEqual(1, len(args.flags))

        flag    =   args.flags[0]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-f1')
        self.assertIsNone(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'f1')
        self.assertEqual(flag.name              ,   '-f1')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '-f1')
        self.assertEqual(flag                   ,   '-f1')

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertTrue(args.options)
        self.assertEqual(1, len(args.options))

        option    =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   3)
        self.assertEqual(option.given_name      ,   '--first-option')
        self.assertIsNone(option.argument_specification)
        self.assertEqual(option.given_hyphens   ,   2)
        self.assertEqual(option.given_label     ,   'first-option')
        self.assertEqual(option.name            ,   '--first-option')
        self.assertEqual(option.value           ,   'val1')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--first-option=val1')
        self.assertEqual(option                 ,   '--first-option=val1')

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertTrue(args.values)
        self.assertEqual(1, len(args.values))

        self.assertEqual('value1', args.values[0])


    def test_double_hyphen_1(self):

        argv    =   ( 'myprog', '-f1', 'value1', '--', '-f2' )
        args    =   clasp.parse(argv)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertTrue(args.flags)
        self.assertEqual(1, len(args.flags))

        flag    =   args.flags[0]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-f1')
        self.assertIsNone(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'f1')
        self.assertEqual(flag.name              ,   '-f1')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '-f1')
        self.assertEqual(flag                   ,   '-f1')

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertFalse(args.options)

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertTrue(args.values)
        self.assertEqual(2, len(args.values))

        self.assertEqual('value1', args.values[0])
        self.assertEqual('-f2', args.values[1])


    def test_double_hyphen_2(self):

        argv    =   ( 'myprog', '-f1', 'value1', '--', '-f2', '--', '--option1=v1' )
        args    =   clasp.parse(argv)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertTrue(args.flags)
        self.assertEqual(1, len(args.flags))

        flag    =   args.flags[0]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-f1')
        self.assertIsNone(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'f1')
        self.assertEqual(flag.name              ,   '-f1')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '-f1')
        self.assertEqual(flag                   ,   '-f1')

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertFalse(args.options)

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertTrue(args.values)
        self.assertEqual(4, len(args.values))

        self.assertEqual('value1', args.values[0])
        self.assertEqual('-f2', args.values[1])
        self.assertEqual('--', args.values[2])
        self.assertEqual('--option1=v1', args.values[3])


    def test_one_flag_and_one_option_and_one_value_with_empty_specifications(self):

        specifications_list    =   ( tuple(), list(), None )

        for specifications in specifications_list:

            argv    =   ( 'myprog', '-f1', 'value1', '--first-option=val1' )
            args    =   clasp.parse(argv, specifications)

            self.assertEqual('myprog', args.program_name)

            self.assertIsInstance(args.flags, ( tuple, ))
            self.assertTrue(args.flags)
            self.assertEqual(1, len(args.flags))

            flag    =   args.flags[0]

            self.assertIsInstance(flag, ( Flag, ))
            self.assertEqual(flag.given_index       ,   1)
            self.assertEqual(flag.given_name        ,   '-f1')
            self.assertIsNone(flag.argument_specification)
            self.assertEqual(flag.given_hyphens     ,   1)
            self.assertEqual(flag.given_label       ,   'f1')
            self.assertEqual(flag.name              ,   '-f1')
            self.assertEqual(flag.extras            ,   {})
            self.assertEqual(str(flag)              ,   '-f1')
            self.assertEqual(flag                   ,   '-f1')

            self.assertIsInstance(args.options, ( tuple, ))
            self.assertTrue(args.options)
            self.assertEqual(1, len(args.options))

            option    =   args.options[0]

            self.assertIsInstance(option, ( Option, ))
            self.assertEqual(option.given_index     ,   3)
            self.assertEqual(option.given_name      ,   '--first-option')
            self.assertIsNone(option.argument_specification)
            self.assertEqual(option.given_hyphens   ,   2)
            self.assertEqual(option.given_label     ,   'first-option')
            self.assertEqual(option.name            ,   '--first-option')
            self.assertEqual(option.value           ,   'val1')
            self.assertEqual(option.extras          ,   {})
            self.assertEqual(str(option)            ,   '--first-option=val1')
            self.assertEqual(option                 ,   '--first-option=val1')

            self.assertIsInstance(args.values, ( tuple, ))
            self.assertTrue(args.values)
            self.assertEqual(1, len(args.values))

            self.assertEqual('value1', args.values[0])


    def test_alias_of_flag_with_one_specification(self):

        flag_verbose    =   clasp.flag('--verbose', alias = '-v', extras = { 'x-name': 'v-val' })

        specifications =   (

            flag_verbose,
        )
        argv    =   ( 'myprog', '--verbose', '--succinct', 'value', '-v' )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertTrue(args.flags)
        self.assertEqual(2, len(args.flags))

        flag    =   args.flags[0]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   4)
        self.assertEqual(flag.given_name        ,   '-v')
        self.assertEqual(flag.argument_specification, flag_verbose)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'v')
        self.assertEqual(flag.name              ,   '--verbose')
        self.assertEqual(flag.extras            ,   { 'x-name': 'v-val' })
        self.assertEqual(str(flag)              ,   '--verbose')
        self.assertEqual(flag                   ,   '--verbose')

        flag    =   args.flags[1]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   2)
        self.assertEqual(flag.given_name        ,   '--succinct')
        self.assertIsNone(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   2)
        self.assertEqual(flag.given_label       ,   'succinct')
        self.assertEqual(flag.name              ,   '--succinct')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '--succinct')
        self.assertEqual(flag                   ,   '--succinct')

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertFalse(args.options)

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertTrue(args.values)
        self.assertEqual(1, len(args.values))

        self.assertEqual('value', args.values[0])


    def alias_of_flag_with_two_specifications(self):

        flag_expand =   clasp.flag('--expand', aliases = ( '-x', '--x', ), extras = { 'some-value': ( 'e', 'x', 't', 'r', 'a', 's', ) })

        specifications =   (

            flag_expand,
        )
        argv    =   ( 'myprog', '-f1', 'value1', '-x', '--delete', '--x', )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertTrue(args.flags)
        self.assertEqual(4, len(args.flags))

        flag    =   args.flags[0]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-f1')
        self.assertIsNone(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'f1')
        self.assertEqual(flag.name              ,   '-f1')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '-f1')
        self.assertEqual(flag                   ,   '-f1')

        flag    =   args.flags[1]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   3)
        self.assertEqual(flag.given_name        ,   '-x')
        self.assertEqual(flag.argument_specification, flag_expand)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'x')
        self.assertEqual(flag.name              ,   '--expand')
        self.assertTrue(flag.extras)
        self.assertEqual(str(flag)              ,   '--expand')
        self.assertEqual(flag                   ,   '--expand')

        flag    =   args.flags[2]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   4)
        self.assertEqual(flag.given_name        ,   '--delete')
        self.assertIsNone(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   2)
        self.assertEqual(flag.given_label       ,   'delete')
        self.assertEqual(flag.name              ,   '--delete')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '--delete')
        self.assertEqual(flag                   ,   '--delete')

        flag    =   args.flags[3]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   5)
        self.assertEqual(flag.given_name        ,   '--x')
        self.assertEqual(flag.argument_specification, specifications[x])
        self.assertEqual(flag.given_hyphens     ,   2)
        self.assertEqual(flag.given_label       ,   'x')
        self.assertEqual(flag.name              ,   '--expand')
        self.assertTrue(flag.extras)
        self.assertIsInstance(flag.extras, dict)
        self.assertEqual(1, len(flag.extras))
        self.assertEqual(( 'e', 'x', 't', 'r', 'a', 's', ), flag.extras['some-value'])
        self.assertEqual(str(flag)              ,   '--expand')
        self.assertEqual(flag                   ,   '--expand')


        self.assertIsInstance(args.options, ( tuple, ))
        self.assertFalse(args.options)

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertTrue(args.values)
        self.assertEqual(1, len(args.values))

        self.assertEqual('value1', args.values[0])


    def test_alias_of_option_with_one_specification(self):

        option_option   =   clasp.option('--option', alias = '-o')

        specifications =   (

            option_option,
        )
        argv    =   ( 'myprog', '-f1', 'value1', '-o=value2', )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertTrue(args.flags)
        self.assertEqual(1, len(args.flags))

        flag    =   args.flags[0]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-f1')
        self.assertIsNone(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'f1')
        self.assertEqual(flag.name              ,   '-f1')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '-f1')
        self.assertEqual(flag                   ,   '-f1')

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertTrue(args.options)
        self.assertEqual(1, len(args.options))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   3)
        self.assertEqual(option.given_name      ,   '-o')
        self.assertEqual(option.argument_specification, option_option)
        self.assertEqual(option.given_hyphens   ,   1)
        self.assertEqual(option.given_label     ,   'o')
        self.assertEqual(option.name            ,   '--option')
        self.assertEqual(option.value           ,   'value2')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--option=value2')
        self.assertEqual(option                 ,   '--option=value2')

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertTrue(args.values)
        self.assertEqual(1, len(args.values))

        self.assertEqual('value1', args.values[0])


    def test_alias_of_option_with_separate_value(self):

        option_option   =   clasp.option('--option', alias = '-o')

        specifications =   (

            option_option,
        )
        argv    =   ( 'myprog', '-o', 'value-1', )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertFalse(args.flags)
        self.assertEqual(0, len(args.flags))

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertTrue(args.options)
        self.assertEqual(1, len(args.options))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   1)
        self.assertEqual(option.given_name      ,   '-o')
        self.assertEqual(option.argument_specification, option_option)
        self.assertEqual(option.given_hyphens   ,   1)
        self.assertEqual(option.given_label     ,   'o')
        self.assertEqual(option.name            ,   '--option')
        self.assertEqual(option.value           ,   'value-1')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--option=value-1')
        self.assertEqual(option                 ,   '--option=value-1')

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertFalse(args.values)
        self.assertEqual(0, len(args.values))


    def test_alias_of_option_that_has_default_with_separate_value(self):

        option_option   =   clasp.option('--option', alias = '-o', default_value = 'def-val-1')

        specifications =   (

            option_option,
        )
        argv    =   ( 'myprog', '-o', 'value-1', )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertFalse(args.flags)
        self.assertEqual(0, len(args.flags))

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertTrue(args.options)
        self.assertEqual(1, len(args.options))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   1)
        self.assertEqual(option.given_name      ,   '-o')
        self.assertEqual(option.argument_specification, option_option)
        self.assertEqual(option.given_hyphens   ,   1)
        self.assertEqual(option.given_label     ,   'o')
        self.assertEqual(option.name            ,   '--option')
        self.assertEqual(option.value           ,   'value-1')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--option=value-1')
        self.assertEqual(option                 ,   '--option=value-1')

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertFalse(args.values)
        self.assertEqual(0, len(args.values))


    def test_alias_of_option_that_has_default_with_separate_value_that_resembles_flag(self):

        option_option   =   clasp.option('--option', alias = '-o', default_value = 'def-val-1')

        specifications =   (

            option_option,
        )
        argv    =   ( 'myprog', '-o', '-o', )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertFalse(args.flags)
        self.assertEqual(0, len(args.flags))

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertTrue(args.options)
        self.assertEqual(1, len(args.options))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   1)
        self.assertEqual(option.given_name      ,   '-o')
        self.assertEqual(option.argument_specification, option_option)
        self.assertEqual(option.given_hyphens   ,   1)
        self.assertEqual(option.given_label     ,   'o')
        self.assertEqual(option.name            ,   '--option')
        self.assertEqual(option.value           ,   '-o')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--option=-o')
        self.assertEqual(option                 ,   '--option=-o')

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertFalse(args.values)
        self.assertEqual(0, len(args.values))


    def test_alias_of_option_that_has_default_with_missing_separate_value(self):

        option_option   =   clasp.option('--option', alias = '-o', default_value = 'def-val-1')

        specifications =   (

            option_option,
        )
        argv    =   ( 'myprog', '-o', )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertFalse(args.flags)
        self.assertEqual(0, len(args.flags))

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertTrue(args.options)
        self.assertEqual(1, len(args.options))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   1)
        self.assertEqual(option.given_name      ,   '-o')
        self.assertEqual(option.argument_specification, option_option)
        self.assertEqual(option.given_hyphens   ,   1)
        self.assertEqual(option.given_label     ,   'o')
        self.assertEqual(option.name            ,   '--option')
        self.assertEqual(option.value           ,   'def-val-1')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--option=def-val-1')
        self.assertEqual(option                 ,   '--option=def-val-1')

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertFalse(args.values)
        self.assertEqual(0, len(args.values))


    def test_alias_of_option_with_attached_empty(self):

        specifications =   (

            clasp.option('--option', alias = '-o', default_value = 'def-val-1'),
        )
        argv    =   ( 'myprog', '-o=', 'value-2', )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertFalse(args.flags)
        self.assertEqual(0, len(args.flags))

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertTrue(args.options)
        self.assertEqual(1, len(args.options))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   1)
        self.assertEqual(option.given_name      ,   '-o')
        self.assertEqual(option.argument_specification  ,   specifications[0])
        self.assertEqual(option.given_hyphens   ,   1)
        self.assertEqual(option.given_label     ,   'o')
        self.assertEqual(option.name            ,   '--option')
        self.assertEqual(option.value           ,   'def-val-1')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--option=def-val-1')
        self.assertEqual(option                 ,   '--option=def-val-1')

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertTrue(args.values)
        self.assertEqual(1, len(args.values))

        self.assertEqual('value-2', args.values[0])


    def test_flag_alias_of_option_with_value(self):

        option_verbosity    =   clasp.option('--verbosity')

        specifications =   (

            option_verbosity,
            clasp.flag('--verbosity=high', alias = '-v'),
        )
        argv    =   ( 'myprog', '-v', )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertFalse(args.flags)
        self.assertEqual(0, len(args.flags))

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertTrue(args.options)
        self.assertEqual(1, len(args.options))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   1)
        self.assertEqual(option.given_name      ,   '-v')
        self.assertEqual(option.argument_specification  ,   specifications[0])
        self.assertEqual(option.given_hyphens   ,   1)
        self.assertEqual(option.given_label     ,   'v')
        self.assertEqual(option.name            ,   '--verbosity')
        self.assertEqual(option.value           ,   'high')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--verbosity=high')
        self.assertEqual(option                 ,   '--verbosity=high')

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertFalse(args.values)
        self.assertEqual(0, len(args.values))


    def test_alias_of_option_with_value_allowing_multiple(self):

        option_option   =   clasp.option('--option', alias = '-o', default_value = 'default-value', on_multiple='allow')

        specifications =   (

            option_option,
        )
        argv    =   ( 'myprog', '-f1', 'value-1', '-o=', '-o=given-value-1', '--option=given-value-2', )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertTrue(args.flags)
        self.assertEqual(1, len(args.flags))


        flag    =   args.flags[0]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-f1')
        self.assertIsNone(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'f1')
        self.assertEqual(flag.name              ,   '-f1')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '-f1')
        self.assertEqual(flag                   ,   '-f1')

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertTrue(args.options)
        self.assertEqual(3, len(args.options))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   3)
        self.assertEqual(option.given_name      ,   '-o')
        self.assertEqual(option.argument_specification  ,   specifications[0])
        self.assertEqual(option.given_hyphens   ,   1)
        self.assertEqual(option.given_label     ,   'o')
        self.assertEqual(option.name            ,   '--option')
        self.assertEqual(option.value           ,   'default-value')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--option=default-value')
        self.assertEqual(option                 ,   '--option=default-value')

        option  =   args.options[1]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   4)
        self.assertEqual(option.given_name      ,   '-o')
        self.assertEqual(option.argument_specification  ,   specifications[0])
        self.assertEqual(option.given_hyphens   ,   1)
        self.assertEqual(option.given_label     ,   'o')
        self.assertEqual(option.name            ,   '--option')
        self.assertEqual(option.value           ,   'given-value-1')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--option=given-value-1')
        self.assertEqual(option                 ,   '--option=given-value-1')

        option  =   args.options[2]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   5)
        self.assertEqual(option.given_name      ,   '--option')
        self.assertEqual(option.argument_specification  ,   specifications[0])
        self.assertEqual(option.given_hyphens   ,   2)
        self.assertEqual(option.given_label     ,   'option')
        self.assertEqual(option.name            ,   '--option')
        self.assertEqual(option.value           ,   'given-value-2')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--option=given-value-2')
        self.assertEqual(option                 ,   '--option=given-value-2')

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertTrue(args.values)
        self.assertEqual(1, len(args.values))

        self.assertEqual('value-1', args.values[0])

    def test_alias_of_option_with_value_ignoring_multiple(self):

        option_option   =   clasp.option('--option', alias = '-o', default_value = 'default-value', on_multiple='ignore')

        specifications =   (

            option_option,
        )
        argv    =   ( 'myprog', '-f1', 'value-1', '-o=', '-o=given-value-1', '--option=given-value-2', )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertTrue(args.flags)
        self.assertEqual(1, len(args.flags))


        flag    =   args.flags[0]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-f1')
        self.assertIsNone(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'f1')
        self.assertEqual(flag.name              ,   '-f1')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '-f1')
        self.assertEqual(flag                   ,   '-f1')

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertTrue(args.options)
        self.assertEqual(3, len(args.options))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   3)
        self.assertEqual(option.given_name      ,   '-o')
        self.assertEqual(option.argument_specification  ,   specifications[0])
        self.assertEqual(option.given_hyphens   ,   1)
        self.assertEqual(option.given_label     ,   'o')
        self.assertEqual(option.name            ,   '--option')
        self.assertEqual(option.value           ,   'default-value')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--option=default-value')
        self.assertEqual(option                 ,   '--option=default-value')

        option  =   args.options[1]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   4)
        self.assertEqual(option.given_name      ,   '-o')
        self.assertEqual(option.argument_specification  ,   specifications[0])
        self.assertEqual(option.given_hyphens   ,   1)
        self.assertEqual(option.given_label     ,   'o')
        self.assertEqual(option.name            ,   '--option')
        self.assertEqual(option.value           ,   'given-value-1')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--option=given-value-1')
        self.assertEqual(option                 ,   '--option=given-value-1')

        option  =   args.options[2]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   5)
        self.assertEqual(option.given_name      ,   '--option')
        self.assertEqual(option.argument_specification  ,   specifications[0])
        self.assertEqual(option.given_hyphens   ,   2)
        self.assertEqual(option.given_label     ,   'option')
        self.assertEqual(option.name            ,   '--option')
        self.assertEqual(option.value           ,   'given-value-2')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--option=given-value-2')
        self.assertEqual(option                 ,   '--option=given-value-2')

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertTrue(args.values)
        self.assertEqual(1, len(args.values))

        self.assertEqual('value-1', args.values[0])

    def test_alias_of_option_with_value_replacing_multiple(self):

        option_option   =   clasp.option('--option', alias = '-o', default_value = 'default-value', on_multiple='replace')

        specifications =   (

            option_option,
        )
        argv    =   ( 'myprog', '-f1', 'value-1', '-o=', '-o=given-value-1', '--option=given-value-2', )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertTrue(args.flags)
        self.assertEqual(1, len(args.flags))


        flag    =   args.flags[0]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-f1')
        self.assertIsNone(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'f1')
        self.assertEqual(flag.name              ,   '-f1')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '-f1')
        self.assertEqual(flag                   ,   '-f1')

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertTrue(args.options)
        self.assertEqual(3, len(args.options))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   3)
        self.assertEqual(option.given_name      ,   '-o')
        self.assertEqual(option.argument_specification  ,   specifications[0])
        self.assertEqual(option.given_hyphens   ,   1)
        self.assertEqual(option.given_label     ,   'o')
        self.assertEqual(option.name            ,   '--option')
        self.assertEqual(option.value           ,   'default-value')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--option=default-value')
        self.assertEqual(option                 ,   '--option=default-value')

        option  =   args.options[1]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   4)
        self.assertEqual(option.given_name      ,   '-o')
        self.assertEqual(option.argument_specification  ,   specifications[0])
        self.assertEqual(option.given_hyphens   ,   1)
        self.assertEqual(option.given_label     ,   'o')
        self.assertEqual(option.name            ,   '--option')
        self.assertEqual(option.value           ,   'given-value-1')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--option=given-value-1')
        self.assertEqual(option                 ,   '--option=given-value-1')

        option  =   args.options[2]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index     ,   5)
        self.assertEqual(option.given_name      ,   '--option')
        self.assertEqual(option.argument_specification  ,   specifications[0])
        self.assertEqual(option.given_hyphens   ,   2)
        self.assertEqual(option.given_label     ,   'option')
        self.assertEqual(option.name            ,   '--option')
        self.assertEqual(option.value           ,   'given-value-2')
        self.assertEqual(option.extras          ,   {})
        self.assertEqual(str(option)            ,   '--option=given-value-2')
        self.assertEqual(option                 ,   '--option=given-value-2')

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertTrue(args.values)
        self.assertEqual(1, len(args.values))

        self.assertEqual('value-1', args.values[0])

    def test_flags_combined(self):

        flag_compile    =   clasp.flag('--compile', alias = '-c')
        flag_debug      =   clasp.flag('--debug', alias = '-d')
        flag_execute    =   clasp.flag('--execute', alias = '-e')

        specifications =   (

            flag_compile,
            flag_debug,
            flag_execute,
        )

        self.assertEqual(flag_compile, specifications[0])
        self.assertEqual(flag_debug, specifications[1])
        self.assertEqual(flag_execute, specifications[2])

        argv    =   ( 'myprog', '-ced', )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertTrue(args.flags)
        self.assertEqual(3, len(args.flags))

        flag    =   args.flags[0]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-ced')
        self.assertEqual(flag.argument_specification, flag_compile)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'ced')
        self.assertEqual(flag.name              ,   '--compile')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '--compile')
        self.assertEqual(flag                   ,   '--compile')
        self.assertEqual(flag, flag_compile)
        self.assertEqual(flag_compile, flag)

        flag    =   args.flags[1]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-ced')
        self.assertEqual(flag.argument_specification, flag_execute)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'ced')
        self.assertEqual(flag.name              ,   '--execute')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '--execute')
        self.assertEqual(flag                   ,   '--execute')
        self.assertEqual(flag, flag_execute)
        self.assertEqual(flag_execute, flag)

        flag    =   args.flags[2]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-ced')
        self.assertEqual(flag.argument_specification, flag_debug)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'ced')
        self.assertEqual(flag.name              ,   '--debug')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '--debug')
        self.assertEqual(flag                   ,   '--debug')
        self.assertEqual(flag_debug, flag)
        self.assertEqual(flag, flag_debug)

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertFalse(args.options)
        self.assertEqual(0, len(args.options))

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertFalse(args.values)
        self.assertEqual(0, len(args.values))


    def test_flags_of_flags_and_options_combined(self):

        flag_compile    =   clasp.flag('--compile', alias = '-c')
        flag_execute    =   clasp.flag('--execute', alias = '-e')
        option_mode     =   clasp.option('--mode', alias = '-m')

        specifications =   (

            flag_compile,
            clasp.flag('--mode=debug', alias = '-d'),
            flag_execute,
            option_mode,
        )

        argv    =   ( 'myprog', '-ced', )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertTrue(args.flags)
        self.assertEqual(2, len(args.flags))

        flag    =   args.flags[0]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-ced')
        self.assertEqual(flag.argument_specification, flag_compile)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'ced')
        self.assertEqual(flag.name              ,   '--compile')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '--compile')
        self.assertEqual(flag                   ,   '--compile')
        self.assertEqual(flag, flag_compile)
        self.assertEqual(flag_compile, flag)

        flag    =   args.flags[1]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-ced')
        self.assertEqual(flag.argument_specification, flag_execute)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'ced')
        self.assertEqual(flag.name              ,   '--execute')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '--execute')
        self.assertEqual(flag                   ,   '--execute')
        self.assertEqual(flag, flag_execute)
        self.assertEqual(flag_execute, flag)

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertTrue(args.options)
        self.assertEqual(1, len(args.options))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index       ,   1)
        self.assertEqual(option.given_name        ,   '-ced')
        self.assertEqual(option.argument_specification, option_mode)
        self.assertEqual(option.given_hyphens     ,   1)
        self.assertEqual(option.given_label       ,   'ced')
        self.assertEqual(option.name              ,   '--mode')
        self.assertEqual(option.extras            ,   {})
        self.assertEqual(str(option)              ,   '--mode=debug')
        self.assertEqual(option                   ,   '--mode=debug')
        self.assertEqual(option_mode, option)
        self.assertEqual(option, option_mode)

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertFalse(args.values)
        self.assertEqual(0, len(args.values))

    def test_first_unused_Flag_via_get_first_unused_flag(self):

        flag_compile    =   clasp.flag('--compile', alias = '-c')
        flag_debug      =   clasp.flag('--debug', alias = '-d')

        specifications = (

            flag_compile,
            flag_debug,
        )

        argv    =   ( 'dir1/myprog', '-cd' )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertTrue(args.flags)
        self.assertEqual(2, len(args.flags))


        # now check the 'unused', iteratively using and testing

        self.assertIsNone(args.get_first_unused_option())

        # before any use()d

        fu      =   args.get_first_unused_flag()

        self.assertIsNotNone(fu)
        self.assertEqual(flag_compile, fu)

        # after use() (1st time)

        fu.use()

        fu      =   args.get_first_unused_flag()

        self.assertIsNotNone(fu)
        self.assertEqual(flag_debug, fu)

        # after use() (2nd time)

        fu.use()

        fu      =   args.get_first_unused_flag()

        self.assertIsNone(fu)

    def test_first_unused_Flag_via_get_first_unused_flag_or_option(self):

        flag_compile    =   clasp.flag('--compile', alias = '-c')
        flag_debug      =   clasp.flag('--debug', alias = '-d')

        specifications = (

            flag_compile,
            flag_debug,
        )

        argv    =   ( 'dir1/myprog', '-cd' )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertTrue(args.flags)
        self.assertEqual(2, len(args.flags))


        # now check the 'unused', iteratively using and testing

        self.assertIsNone(args.get_first_unused_option())

        # before any use()d

        fu      =   args.get_first_unused_flag_or_option()

        self.assertIsNotNone(fu)
        self.assertEqual(flag_compile, fu)

        # after use() (1st time)

        fu.use()

        fu      =   args.get_first_unused_flag_or_option()

        self.assertIsNotNone(fu)
        self.assertEqual(flag_debug, fu)

        # after use() (2nd time)

        fu.use()

        fu      =   args.get_first_unused_flag_or_option()

        self.assertIsNone(fu)

    def test_first_unused_Flag_via_get_first_unused(self):

        flag_compile    =   clasp.flag('--compile', alias = '-c')
        flag_debug      =   clasp.flag('--debug', alias = '-d')

        specifications = (

            flag_compile,
            flag_debug,
        )

        argv    =   ( 'dir1/myprog', '-cd' )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertTrue(args.flags)
        self.assertEqual(2, len(args.flags))


        # now check the 'unused', iteratively using and testing

        self.assertIsNone(args.get_first_unused_option())

        # before any use()d

        fu      =   args.get_first_unused()

        self.assertIsNotNone(fu)
        self.assertEqual(flag_compile, fu)

        # after use() (1st time)

        fu.use()

        fu      =   args.get_first_unused()

        self.assertIsNotNone(fu)
        self.assertEqual(flag_debug, fu)

        # after use() (2nd time)

        fu.use()

        fu      =   args.get_first_unused()

        self.assertIsNone(fu)

    def test_first_unused_Option_via_get_first_unused_option(self):

        option_mode     =   clasp.option('--mode', alias = '-m')
        option_option   =   clasp.option('--option', alias = '-o', default_value = 'default-value', on_multiple='replace')

        specifications = (

            option_mode,
            option_option,
        )

        argv    =   ( 'dir1/myprog', '--mode=verbose', '--option=ignore' )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertTrue(args.options)
        self.assertEqual(2, len(args.options))


        # now check the 'unused', iteratively using and testing

        self.assertIsNone(args.get_first_unused_flag())

        # before any use()d

        fu      =   args.get_first_unused_option()

        self.assertIsNotNone(fu)
        self.assertEqual(option_mode, fu)

        # after use() (1st time)

        fu.use()

        fu      =   args.get_first_unused_option()

        self.assertIsNotNone(fu)
        self.assertEqual(option_option, fu)

        # after use() (2nd time)

        fu.use()

        fu      =   args.get_first_unused_option()

        self.assertIsNone(fu)

    def test_first_unused_Option_via_get_first_unused_flag_or_option(self):

        option_mode     =   clasp.option('--mode', alias = '-m')
        option_option   =   clasp.option('--option', alias = '-o', default_value = 'default-value', on_multiple='replace')

        specifications = (

            option_mode,
            option_option,
        )

        argv    =   ( 'dir1/myprog', '--mode=verbose', '--option=ignore' )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertTrue(args.options)
        self.assertEqual(2, len(args.options))


        # now check the 'unused', iteratively using and testing

        self.assertIsNone(args.get_first_unused_flag())

        # before any use()d

        fu      =   args.get_first_unused_flag_or_option()

        self.assertIsNotNone(fu)
        self.assertEqual(option_mode, fu)

        # after use() (1st time)

        fu.use()

        fu      =   args.get_first_unused_flag_or_option()

        self.assertIsNotNone(fu)
        self.assertEqual(option_option, fu)

        # after use() (2nd time)

        fu.use()

        fu      =   args.get_first_unused_flag_or_option()

        self.assertIsNone(fu)

    def test_first_unused_Option_via_get_first_unused(self):

        option_mode     =   clasp.option('--mode', alias = '-m')
        option_option   =   clasp.option('--option', alias = '-o', default_value = 'default-value', on_multiple='replace')

        specifications = (

            option_mode,
            option_option,
        )

        argv    =   ( 'dir1/myprog', '--mode=verbose', '--option=ignore' )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertTrue(args.options)
        self.assertEqual(2, len(args.options))


        # now check the 'unused', iteratively using and testing

        self.assertIsNone(args.get_first_unused_flag())

        # before any use()d

        fu      =   args.get_first_unused()

        self.assertIsNotNone(fu)
        self.assertEqual(option_mode, fu)

        # after use() (1st time)

        fu.use()

        fu      =   args.get_first_unused()

        self.assertIsNotNone(fu)
        self.assertEqual(option_option, fu)

        # after use() (2nd time)

        fu.use()

        fu      =   args.get_first_unused()

        self.assertIsNone(fu)



if '__main__' == __name__:

    unittest.main()





#!/usr/bin/env python

import pyclasp as clasp

import sys

specifications = (


    clasp.flag('--version', alias = '-v', help = 'show the program version and quits'),

    clasp.option('--verbosity', help = 'the verbosity', values = [ 'silent', 'quiet', 'succinct', 'chatty', 'verbose' ]),
    clasp.option('--length', alias = '-l', help = 'specifies the length'),
    clasp.flag('--verbosity=succinct', aliases = [ '--succinct', '-s' ]),
    clasp.flag('--verbosity=verbose', alias = '--verbose'),
)

INFO_LINES = (

    'CLASP.Python examples',
    ':version',
    "Illustrates use of CLASP.Python's clasp.show_usage() and clasp.show_version() methods",
    '',
)

args = clasp.parse(sys.argv, specifications)

print('*' * 50 + "\n")
print("usage:\n")
clasp.show_usage(args, version = [ 1, 2, 3 ], stream = sys.stdout, program_name = 'myprog', version_prefix = 'v', info_lines=INFO_LINES)

print('*' * 50 + "\n")
print("version:\n")
clasp.show_version(args, version = [ 1, 2, 3 ], stream = sys.stdout, program_name = 'myprog', version_prefix = 'v')

print





from .specification import Specification
from .flag_specification import FlagSpecification
from .flag_argument import FlagArgument
from .option_specification import OptionSpecification
from .option_argument import OptionArgument

from .util import _get_program_name
from .util import _global_multiple_flags_action, _global_multiple_options_action
from .util import _MULTIPLE_ACTION_ALLOW, _MULTIPLE_ACTION_IGNORE, _MULTIPLE_ACTION_REJECT, _MULTIPLE_ACTION_REPLACE

import re
import sys


class Arguments:
    """
    Represents a parsed command-line, separated into program-name, flags, options, and values
    """


    def __init__(self, argv, specifications = None):
        """
        Initialises an instance from the given `argv` and `specifications` sequences. Users should instead use `clasp.parse()`
        """

        if not isinstance(argv, ( list, tuple )):

            raise TypeError("`argv` argument must be an instance of `list` or `tuple`")

        if specifications and not isinstance(specifications, ( list, tuple )):

            raise TypeError("`specifications` argument must be `None` or an instance of `list` or `tuple`")

        self.argv       =   argv

        self.specifications     =   specifications if specifications else ()

        flags, options, values  =   Arguments._parse(argv, self.specifications)

        self.program_name       =   Arguments.get_program_name(argv)
        """
        str
            The program name
        """

        self.flags      =   tuple(flags)
        """
        tuple
            The parsed flags
        """

        self.options    =   tuple(options)
        """
        tuple
            The parsed options
        """

        self.values     =   tuple(values)
        """
        tuple
            The parsed values
        """


    def aliases(self):
        """
        [DEPRECATED] instead use `specifications`
        """

        return self.specifications


    def flagIsSpecified(self, id):
        """
        [DEPRECATED] Use `flag_is_specified()`
        """

        return self.flag_is_specified(id)


    def flag_is_specified(self, id):
        """
        Returns
        -------
        bool
            `True` if the given flag (name, or instance) has been specified; `False` otherwise
        """

        return None != self.lookup_flag(id);


    def lookupFlag(self, id):
        """
        [DEPRECATED] Use `lookup_flag()`
        """

        return self.lookup_flag(id)


    def lookup_flag(self, id):
        """
        Looks and returns the identified flag from the instance's flags; returns `None` if not found
        """

        name    =   None

        if False:

            pass
        elif FlagSpecification == type(id):

            name    =   id.name
        elif OptionSpecification == type(id):

            name    =   id.name
        else:

            name    =   id

        for index, flag in enumerate(self.flags):

            if flag.name == name:

                flag.use()

                return flag

        return None


    def lookupOption(self, id):
        """
        [DEPRECATED] Use `lookup_option()`
        """

        return self.lookup_option(id)


    def lookup_option(self, id):
        """
        Looks and returns the identified option from the instance's options; returns `None` if not found
        """

        name    =   None

        if False:

            pass
        elif FlagSpecification == type(id):

            name    =   id.name
        elif OptionSpecification == type(id):

            name    =   id.name
        else:

            name    =   id

        for index, option in enumerate(self.options):

            if option.name == name:

                option.use()

                return option

        return None


    def getFirstUnusedFlag(self):
        """
        [DEPRECATED] Use `get_first_unused_flag()`
        """

        return self.get_first_unused_flag()


    def get_first_unused_flag(self, id=None):
        """
        Looks and returns the first unused flag from the instance's flags; returns `None` if not found

        If the argument `id` is specified, only matching unused flags will be obtained
        """

        for flag in self.flags:

            if id and not Arguments._id_matches(flag, id):

                continue

            if not flag.used():

                return flag

        return None


    def getFirstUnusedOption(self):
        """
        [DEPRECATED] Use `get_first_unused_option()`
        """

        return self.get_first_unused_option()


    def get_first_unused_option(self, id=None):
        """
        Looks and returns the first unused option from the instance's options; returns `None` if not found

        If the argument `id` is specified, only matching unused options will be obtained
        """

        for option in self.options:

            if id and not Arguments._id_matches(option, id):

                continue

            if not option.used():

                return option

        return None


    def getFirstUnusedFlagOrOption(self):
        """
        [DEPRECATED] Use `get_first_unused_flag_or_option()`
        """

        return self.get_first_unused_flag_or_option()


    def get_first_unused(self, id=None):
        """
        Shorthand for `get_first_unused_flag_or_option()`
        """

        return self.get_first_unused_flag_or_option(id)


    def get_first_unused_flag_or_option(self, id=None):
        """
        Obtains a reference to the first unused flag or option, or `None` if all no unused are found

        If the argument `id` is specified, only matching unused flags/options will be obtained

        Returns
        -------
        FlagArgument, OptionArgument, None
            the first unused flag or option, if any
        """

        flag    =   self.get_first_unused_flag(id)
        option  =   self.get_first_unused_option(id)

        if flag:

            if option:

                if flag.given_index < option.given_index:

                    return flag
                else:

                    return option
            else:

                return flag
        else:

            return option


    @staticmethod
    def _id_matches(flag_or_option, id):

        name    =   None

        if False:

            pass
        elif FlagSpecification == type(id):

            name    =   id.name
        elif OptionSpecification == type(id):

            name    =   id.name
        else:

            name    =   id

        return name == flag_or_option.name


    @staticmethod
    def get_program_name(argv=None):
        """
        Obtains/infers the program name from the given array, or from `sys.argv`

        Returns
        -------
        str
            the program name
        """

        if argv is None:

            argv = sys.argv

        return _get_program_name(argv, {})


    @staticmethod
    def _select_specification(item, specifications):

        for spec in specifications:

            if spec.name == item:

                return spec

            for a2 in spec.aliases:

                if item == a2:

                    return spec

        return None


    @staticmethod
    def _select_specifications(item, specifications):

        select_specification = Arguments._select_specification(item, specifications)

        if select_specification:

            return select_specification

        n = len(item) - 1

        if n > 1:

            m = re.match(r'-+([a-zA-Z]+)$', item)

            if m:

                select_specifications  =   []

                for c in m.group(1):

                    name = '-' + c

                    select_specification = Arguments._select_specification(name, specifications)

                    if select_specification:

                        m2 = re.match(r'(-+)([^=]+)=(.*)$', select_specification.name)

                        if m2:

                            name2 = m2.group(1) + m2.group(2)

                            select_option_specification = Arguments._select_specification(name2, specifications)

                            if select_option_specification:

                                select_specifications.append((select_option_specification, select_specification, m2.group(3)))
                        else:

                            select_specifications.append(select_specification)


                if len(select_specifications) == n:

                    return select_specifications

        return None


    @staticmethod
    def _add_flag(flags, flag, spec):

        for i, f in enumerate(flags):

            if f == flag:

                default_action  =   _global_multiple_flags_action()

                if False:

                    pass
                elif _MULTIPLE_ACTION_IGNORE == default_action:

                    return
                elif _MULTIPLE_ACTION_REJECT == default_action:

                    raise DuplicateFlagSpecified(f, flag, spec)
                elif _MULTIPLE_ACTION_REPLACE == default_action:

                    flags[i] = flag

                    return

        flags.append(flag)


    @staticmethod
    def _add_option(options, option, spec):

        for i, o in enumerate(options):

            if o == option:

                action  =   None

                if not spec:

                    spec    =   option.argument_specification

                if spec:

                    action  =   spec.on_multiple
                else:

                    action  =   _global_multiple_options_action()

                if False:

                    pass
                elif _MULTIPLE_ACTION_ALLOW == action:

                    options.append(option)

                    return
                elif _MULTIPLE_ACTION_IGNORE == action:

                    return
                elif _MULTIPLE_ACTION_REJECT == action:

                    raise DuplicateOptionSpecified(o, option, spec)
                elif _MULTIPLE_ACTION_REPLACE == action:

                    options[i] = option

                    return

        options.append(option)


    @staticmethod
    def _parse(argv, specifications):

        flags           =   []
        options         =   []
        values          =   []

        forced_value    =   False
        current_option  =   None

        for index, arg in enumerate(argv):

            if 0 == index:

                # ignore program name
                continue

            if not forced_value:

                if '--' == arg:

                    forced_value = True

                    continue

            if current_option:

                current_option._set_value(arg)

                Arguments._add_option(options, current_option, None)

                current_option          =   None

                continue

            if forced_value:

                values.append(arg)

                continue

            m = re.match(r'(-+)([^=]+)', arg)

            if m:

                hyphens         =   m.group(1)
                given_label     =   m.group(2)
                given_name      =   hyphens + given_label
                resolved_name   =   given_name
                arg_spec        =   None
                extras          =   None
                value           =   None
                is_option       =   False

                gr              =   m.group()

                if gr != arg:

                    # The option has an attached value

                    value       =   arg[1 + len(gr):]
                    is_option   =   True
                else:

                    # an option name, or a flag, or a combined set of flags

                    sel_specifications =   Arguments._select_specifications(arg, specifications)

                    if sel_specifications:

                        if isinstance(sel_specifications, (Specification, )):

                            pass
                        else:

                            # We have a combination of items

                            for spec in sel_specifications:

                                if False:

                                    pass
                                elif isinstance(spec, (FlagSpecification, )):

                                    flag = FlagArgument(arg, index, arg, spec.name, spec, len(hyphens), given_label, spec.extras)

                                    Arguments._add_flag(flags, flag, spec)
                                elif isinstance(spec, (OptionSpecification, )):

                                    pass
                                elif isinstance(spec, (tuple, )):

                                    soa =   spec[0]
                                    fa  =   spec[1]
                                    v   =   spec[2]

                                    option = OptionArgument(arg, index, arg, soa.name, soa, len(hyphens), given_label, v, extras)

                                    Arguments._add_option(options, option, soa)
                                else:

                                    pass

                            continue

                # Now look through the specifications, for:
                #
                # - the resolved name, and
                # - the default value, if none was attached
                for i, spec in enumerate(specifications):

                    if spec.name == given_name or given_name in spec.aliases:

                        is_option       =   isinstance(spec, OptionSpecification)

                        resolved_name   =   spec.name
                        arg_spec        =   spec
                        extras          =   spec.extras

                        hyphens_2       =   None
                        given_label_2   =   None
                        value_2         =   None
                        resolved_name_2 =   None

                        alias_has_value =   False

                        m2 = re.match(r'(-+)([^=]+)=(.*)', resolved_name)

                        # need to check whether the alias is a
                        # valued-option, and, if so, expand out its name
                        # and value, and replace the name and (if not
                        # previously specified) the value

                        if m2:

                            alias_has_value =   True

                            hyphens_2       =   m2.group(1)
                            given_label_2   =   m2.group(2)
                            value_2         =   m2.group(3)
                            resolved_name_2 =   hyphens_2 + given_label_2

                            resolved_name   =   resolved_name_2

                            # now find the underlying (option) specification

                            for j, s2 in enumerate(specifications):

                                if s2.name == resolved_name or resolved_name in s2.aliases:

                                    arg_spec = s2

                                    break

                        if is_option:

                            if not value and value != None:

                                if alias_has_value:

                                    value   =   value_2
                                else:

                                    if spec.default_value:

                                        value   =   spec.default_value
                        else:

                            if alias_has_value:

                                is_option   =   True
                                value       =   value_2

                        break

                if is_option:

                    option      =   OptionArgument(arg, index, given_name, resolved_name, arg_spec, len(hyphens), given_label, value, extras)

                    if value:

                        Arguments._add_option(options, option, arg_spec)
                    else:

                        current_option  =   option
                else:

                    flag        =   FlagArgument(arg, index, given_name, resolved_name, arg_spec, len(hyphens), given_label, extras)

                    Arguments._add_flag(flags, flag, arg_spec)
            else:

                values.append(arg)

        if current_option:

            value   =   None
            spec    =   current_option.argument_specification

            if spec:

                # always push, so value type processing can be done (for missing values)

                current_option._set_value(spec.default_value)

            Arguments._add_option(options, current_option, spec)


        return flags, options, values





class CLASPException(RuntimeError):
    """
    Root exception for CLASP
    """

    pass


class ParsingException(CLASPException):
    """
    Root exception for parsing
    """

    pass


class ValueParsingException(ParsingException):
    """
    Root exception for value parsing
    """

    pass


class MissingValueException(ValueParsingException):
    """
    Used to indicate that no value is specified for an option (and its specification, if any, does not have a default)
    """

    pass


class InvalidValueException(ValueParsingException):
    """
    Root exception for invalid values
    """

    pass


class InvalidBooleanException(InvalidValueException):
    """
    The given value could not be recognised as a (properly-formatted) boolean
    """

    pass


class InvalidNumberException(InvalidValueException):
    """
    The given value could not be recognised as a (properly-formatted) number
    """

    pass


class InvalidIntegerException(InvalidNumberException):
    """
    The given value could not be recognised as a (properly-formatted) integer
    """

    pass


class IntegerOutOfRangeException(InvalidValueException):
    """
    The given value as a valid integer but is out of range
    """

    pass


class DuplicateFlagSpecified(ParsingException):
    """
    Used to indicate that a duplicate flag is specified
    """


    def __init__(self, existing, current, spec):

        self.existing   =   existing
        """
        The existing flag that is duplicated
        """

        self.current    =   current
        """
        The new flag that is a duplicate
        """

        self.specification  =   spec
        """
        The specification of the flag, if any
        """


class DuplicateOptionSpecified(ParsingException):
    """
    Used to indicate that a duplicate option is specified
    """


    def __init__(self, existing, current, spec):

        self.existing   =   existing
        """
        The existing option that is duplicated
        """

        self.current    =   current
        """
        The new option that is a duplicate
        """

        self.specification  =   spec
        """
        The specification of the option, if any
        """




#! /usr/bin/env python

# examples/multiple_options.py


# imports

import pyclasp as clasp

import sys

# constants

VERSION = [ 0, 0, 1 ]

INFO_LINES = (

    'CLASP.Python examples',
    ':version:',
    "Illustrates use of CLASP.Python's support for multiple specification of options",
    '',
)

# Specify specifications, parse, and checking standard flags

option_Extension = clasp.option('--extension', alias='-e', help='specifies a file extension. May be specified multiple times', on_multiple='allow')
flag_CalcDirSize = clasp.flag('--calc-dir-size', alias='-z', help='causes directory sizes to be calculated')

specifications = (

    option_Extension,
    flag_CalcDirSize,

    clasp.HelpFlag(),
    clasp.VersionFlag(),
)

args = clasp.parse(sys.argv, specifications)

if args.flag_is_specified('--help'):

    clasp.show_usage(specifications, exit_code=0, version=VERSION, stream=sys.stdout, info_lines=INFO_LINES)

if args.flag_is_specified('--version'):

    clasp.show_version(specifications, exit_code=0, version=VERSION, stream=sys.stdout)


# List all flags and options for your edification

for f in args.flags:

    print("flag: %s" % f)

    f.use()

for o in args.options:

    print("option: %s" % o)

    o.use()



# Check for any unrecognised flags or options

unused = args.get_first_unused_flag_or_option()
if (unused):

    sys.stderr.write("%s: unrecognised flag/option: %s\n" % (args.program_name, unused))

    sys.exit(1)





#!/usr/bin/env python

import pyclasp as clasp

import unittest

try:

    from StringIO import StringIO
except ImportError:

    from io import StringIO


class Util_tester(unittest.TestCase):

    def test_show_version_1(self):

        stm     =   StringIO()

        try:
            specifications =   ()

            clasp.show_version(specifications, version = "1.2.3", stream = stm, program_name = 'myprog', version_prefix = 'v')

            self.assertEqual('myprog v1.2.3', stm.getvalue().strip())

        finally:
            stm.close()


    def test_show_version_2(self):

        stm     =   StringIO()

        try:
            specifications =   ()

            clasp.show_version(specifications, version = [ 1, 2, 3 ], stream = stm, program_name = 'myprog', version_prefix = 'v')

            self.assertEqual('myprog v1.2.3', stm.getvalue().strip())

        finally:
            stm.close()


    def test_show_version_3(self):

        stm     =   StringIO()

        try:
            specifications =   ()

            clasp.show_version(specifications, version = "1.2.3", stream = stm, program_name = 'myprog')

            self.assertEqual('myprog 1.2.3', stm.getvalue().strip())

        finally:
            stm.close()




if '__main__' == __name__:

    unittest.main()






__author__      =   'Matt Wilson'
__copyright__   =   'Copyright 2019-2020 Synesis Information Systems, Copyright 2019 Synesis Software'
__credits__     =   [

        'Garth Lancaster',
        'Matt Wilson',
 ]
__email__       =   'matthew@synesis.com.au'
__license__     =   'BSD-3-Clause'
__maintainer__  =   'Matt Wilson'
__status__      =   'Beta'
__version__     =   '0.8.9'

from .exceptions import *
from .flag_specification import FlagSpecification, flag
from .flag_specification import HelpFlag, VersionFlag
from .option_specification import OptionSpecification, option
from .section_specification import SectionSpecification, section
from .specification import Specification, specification

from .arguments import Arguments
from .flag_argument import FlagArgument as Flag
from .option_argument import OptionArgument as Option

from .cli import show_usage, show_version

import sys

def parse(argv = None, specifications = None):
    """
    Obtains an instance of `clasp.Arguments`, representing all the command-line arguments present in `argv` (or `sys.argv`)
    """

    if argv is None:

        argv = sys.argv

    return Arguments(argv, specifications)

def get_program_name(argv = None):
    """
    Obtains/infers the program name from the given array, or from `sys.argv`
    """

    return Arguments.get_program_name(argv)




#!/usr/bin/env python

import pyclasp as clasp

import sys

# constants

VERSION = '0.0.2'

INFO_LINES = (

    'CLASP.Python examples',
    ':version',
    "Illustrates use of CLASP.Python's typed option values",
    '',
)

# Specify specifications, parse, and checking standard flags

specifications = (

    clasp.HelpFlag(),
    clasp.VersionFlag(),

    clasp.option('--length', alias = '-l', help = 'specifies the length', value_type=int),
)

try:

    args = clasp.parse(sys.argv, specifications)
except clasp.ValueParsingException as x:

    sys.stderr.write("%s: invalid command line: %s\n" % (clasp.get_program_name(), x))

    sys.exit(1)


if args.flag_is_specified('--help'):

    clasp.show_usage(specifications, exit_code=0, version=VERSION, stream=sys.stdout, info_lines=INFO_LINES)

if args.flag_is_specified('--version'):

    clasp.show_version(specifications, exit_code=0, version=VERSION, stream=sys.stdout)


# Program-specific processing of flags/options

opt_length = args.lookup_option('--length')
if opt_length:

    print("You specified length with the value: %d (of type `%s`). The string that was passed is available in the `given_value` attribute, which is '%s' (of type `%s`)\n" % (opt_length.value, type(opt_length.value), opt_length.given_value, type(opt_length.given_value)))
else:

    sys.stderr.write("try specifying the '--length' option; use --help for usage\n")

    sys.exit(1)



# Check for any unrecognised flags or options

unused = args.get_first_unused_flag_or_option();
if (unused):

    sys.stderr.write("%s: unrecognised flag/option: %s\n" % (args.program_name, unused))

    sys.exit(1)






from .specification import Specification

class FlagSpecification(Specification):
    """
    TBC
    """


    def __init__(self, name, aliases, help, extras):

        super(FlagSpecification, self).__init__(name, aliases, help, extras)


    def __str__(self):

        return "<%s.%s: name=%s; help=%s; aliases=%s; extras=%s>" %\
            (self.__module__, self.__class__.__name__, self.name, self.help, self.aliases, self.extras)


def flag(name, **kwargs):
    """
    Creates a flag specification from the given parameters
    """

    aliases =   None
    help    =   None
    extras  =   None

    for n, v in kwargs.items():

        if False:

            pass
        elif 'alias' == n:

            aliases = ( v, )
        elif 'aliases' == n:

            aliases = v
        elif 'help' == n:

            help = v
        elif 'extras' == n:

            extras = v
        else:

            raise TypeError("`flag` method does not recognise the `%s` keyword argument" % (n, ))

    return FlagSpecification(name, aliases, help, extras)


_HELP_FLAG      =   FlagSpecification('--help', None, 'Shows usage and terminates', None)
_VERSION_FLAG   =   FlagSpecification('--version', None, 'Shows version and terminates', None)


def HelpFlag():

   return _HELP_FLAG


def VersionFlag():

   return _VERSION_FLAG





#!/usr/bin/env python

import pyclasp as clasp

import unittest

import sys

try:

    from StringIO import StringIO
except ImportError:

    from io import StringIO


def _stripped_non_blank_lines_from_SIO(sio):

    s = sio.getvalue().strip()

    lines = s.split("\n")

    lines = [line.strip() for line in lines]

    lines = [line for line in lines if 0 != len(line)]

    return lines

class Sections_tester(unittest.TestCase):

    def test_sections_1(self):

        stm     =   StringIO()

        specs   =   (


            clasp.section('Behaviour:'),

            clasp.flag('--verbose', alias='-v', help='Make output verbose'),

            clasp.section('Standard:'),

            clasp.HelpFlag(),
            clasp.VersionFlag(),
        )

        try:
            clasp.show_usage(specs, exit_code=None, stream=stm, info_lines=(), program_name='myprog')

            actual = _stripped_non_blank_lines_from_SIO(stm)

            expected = (

                'USAGE: myprog [ ... flags and options ... ]',
                'flags/options:',
                'Behaviour:',
                '-v',
                '--verbose',
                'Make output verbose',
                'Standard:',
                '--help',
                'Shows usage and terminates',
                '--version',
                'Shows version and terminates',
            )

            self.assertMultiLineEqual(\
                "\n".join(expected),
                "\n".join(actual),
            )

        finally:
            stm.close()




if '__main__' == __name__:

    unittest.main()







#!/usr/bin/env python

# examples/flags_and_options_specifications.py


# imports

import pyclasp as clasp

import sys

# constants

VERSION = [ 0, 0, 3 ]

INFO_LINES = (

    'CLASP.Python examples',
    ':version:',
    "Illustrates use of CLASP.Python's use of flags, options, and specifications",
    '',
)

# Specify specifications, parse, and checking standard flags

flag_Debug = clasp.flag('--debug', alias='-d', help='runs in Debug mode')
option_Verbosity = clasp.option('--verbosity', alias='-v', help='specifies the verbosity', values=[ 'terse', 'quiet', 'silent', 'chatty' ])
flag_Chatty = clasp.flag('--verbosity=chatty', alias='-c')

specifications = (

    flag_Debug,
    option_Verbosity,
    flag_Chatty,

    clasp.HelpFlag(),
    clasp.VersionFlag(),
)

args = clasp.parse(sys.argv, specifications)

if args.flag_is_specified('--help'):

    clasp.show_usage(specifications, exit_code=0, version=VERSION, stream=sys.stdout, info_lines=INFO_LINES)

if args.flag_is_specified('--version'):

    clasp.show_version(specifications, exit_code=0, version=VERSION, stream=sys.stdout)


# Program-specific processing of flags/options

opt = args.lookup_option('--verbosity')
if (opt):

    sys.stdout.write("verbosity is specified as: %s\n" % opt.value)


if (args.flag_is_specified('--debug')):

    sys.stdout.write("Debug mode is specified\n")


# Check for any unrecognised flags or options

unused = args.get_first_unused_flag_or_option()
if (unused):

    sys.stderr.write("%s: unrecognised flag/option: %s\n" % (args.program_name, unused))

    sys.exit(1)





#!/usr/bin/env python

from pyclasp import Arguments
from pyclasp import specification, option
from pyclasp import Flag
from pyclasp import Option
import pyclasp as clasp

import unittest

import sys

class Typed_values_tester_1(unittest.TestCase):

    def test_valid_option_value_of_type_str(self):

        specifications =   (

            clasp.option('--length', alias='-l', value_type=str),
        )

        argv    =   ( 'myprog', '--length=1.23', '-l', '4.56', )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual(0, len(args.flags))
        self.assertEqual(2, len(args.options))
        self.assertEqual(0, len(args.values))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index       ,   1)
        self.assertEqual(option.given_name        ,   '--length')
        self.assertTrue(option.argument_specification)
        self.assertEqual(option.given_hyphens     ,   2)
        self.assertEqual(option.given_label       ,   'length')
        self.assertEqual(option.name              ,   '--length')
        self.assertEqual(option.extras            ,   {})
        self.assertEqual(option.given_value       ,   '1.23')
        self.assertIsInstance(option.value        ,   str)
        self.assertEqual(str(option)              ,   '--length=1.23')
        self.assertEqual(option                   ,   '--length=1.23')

        option  =   args.options[1]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index       ,   2)
        self.assertEqual(option.given_name        ,   '-l')
        self.assertTrue(option.argument_specification)
        self.assertEqual(option.given_hyphens     ,   1)
        self.assertEqual(option.given_label       ,   'l')
        self.assertEqual(option.name              ,   '--length')
        self.assertEqual(option.extras            ,   {})
        self.assertEqual(option.given_value       ,   '4.56')
        self.assertIsInstance(option.value        ,   str)
        self.assertEqual(str(option)              ,   '--length=4.56')
        self.assertEqual(option                   ,   '--length=4.56')


    def test_valid_option_value_of_type_bool(self):

        specifications =   (

            clasp.option('--verbose', alias='-v', value_type=bool),
        )

        argv    =   ( 'myprog', '--verbose=true', '-v', 'FALSE', )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual(0, len(args.flags))
        self.assertEqual(2, len(args.options))
        self.assertEqual(0, len(args.values))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index       ,   1)
        self.assertEqual(option.given_name        ,   '--verbose')
        self.assertTrue(option.argument_specification)
        self.assertEqual(option.given_hyphens     ,   2)
        self.assertEqual(option.given_label       ,   'verbose')
        self.assertEqual(option.name              ,   '--verbose')
        self.assertEqual(option.extras            ,   {})
        self.assertEqual(option.given_value       ,   'true')
        self.assertIsInstance(option.value        ,   bool)
        self.assertEqual(option.value             ,   True)
        self.assertEqual(str(option)              ,   '--verbose=true')
        self.assertEqual(option                   ,   '--verbose=true')

        option  =   args.options[1]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index       ,   2)
        self.assertEqual(option.given_name        ,   '-v')
        self.assertTrue(option.argument_specification)
        self.assertEqual(option.given_hyphens     ,   1)
        self.assertEqual(option.given_label       ,   'v')
        self.assertEqual(option.name              ,   '--verbose')
        self.assertEqual(option.extras            ,   {})
        self.assertEqual(option.given_value       ,   'FALSE')
        self.assertIsInstance(option.value        ,   bool)
        self.assertEqual(option.value             ,   False)
        self.assertEqual(str(option)              ,   '--verbose=false')
        self.assertEqual(option                   ,   '--verbose=false')


    def test_valid_option_value_of_type_float(self):

        specifications =   (

            clasp.option('--length', alias='-l', value_type=float),
        )

        argv    =   ( 'myprog', '--length=1.23', '-l', '4.56', )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual(0, len(args.flags))
        self.assertEqual(2, len(args.options))
        self.assertEqual(0, len(args.values))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index       ,   1)
        self.assertEqual(option.given_name        ,   '--length')
        self.assertTrue(option.argument_specification)
        self.assertEqual(option.given_hyphens     ,   2)
        self.assertEqual(option.given_label       ,   'length')
        self.assertEqual(option.name              ,   '--length')
        self.assertEqual(option.extras            ,   {})
        self.assertEqual(option.given_value       ,   '1.23')
        self.assertIsInstance(option.value        ,   float)
        self.assertAlmostEqual(option.value       ,   1.23)
        self.assertEqual(str(option)              ,   '--length=1.23')
        self.assertEqual(option                   ,   '--length=1.23')

        option  =   args.options[1]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index       ,   2)
        self.assertEqual(option.given_name        ,   '-l')
        self.assertTrue(option.argument_specification)
        self.assertEqual(option.given_hyphens     ,   1)
        self.assertEqual(option.given_label       ,   'l')
        self.assertEqual(option.name              ,   '--length')
        self.assertEqual(option.extras            ,   {})
        self.assertEqual(option.given_value       ,   '4.56')
        self.assertIsInstance(option.value        ,   float)
        self.assertAlmostEqual(option.value       ,   4.56)
        self.assertEqual(str(option)              ,   '--length=4.56')
        self.assertEqual(option                   ,   '--length=4.56')


    def test_valid_option_value_of_type_int(self):

        specifications =   (

            clasp.option('--length', alias='-l', value_type=int),
        )

        argv    =   ( 'myprog', '--length=123', '-l', '456', )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual(0, len(args.flags))
        self.assertEqual(2, len(args.options))
        self.assertEqual(0, len(args.values))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index       ,   1)
        self.assertEqual(option.given_name        ,   '--length')
        self.assertTrue(option.argument_specification)
        self.assertEqual(option.given_hyphens     ,   2)
        self.assertEqual(option.given_label       ,   'length')
        self.assertEqual(option.name              ,   '--length')
        self.assertEqual(option.extras            ,   {})
        self.assertEqual(option.given_value       ,   '123')
        self.assertIsInstance(option.value        ,   int)
        self.assertEqual(option.value             ,   123)
        self.assertEqual(str(option)              ,   '--length=123')
        self.assertEqual(option                   ,   '--length=123')

        option  =   args.options[1]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index       ,   2)
        self.assertEqual(option.given_name        ,   '-l')
        self.assertTrue(option.argument_specification)
        self.assertEqual(option.given_hyphens     ,   1)
        self.assertEqual(option.given_label       ,   'l')
        self.assertEqual(option.name              ,   '--length')
        self.assertEqual(option.extras            ,   {})
        self.assertEqual(option.given_value       ,   '456')
        self.assertIsInstance(option.value        ,   int)
        self.assertEqual(option.value             ,   456)
        self.assertEqual(str(option)              ,   '--length=456')
        self.assertEqual(option                   ,   '--length=456')


    if sys.version_info < (3, 0):

        def test_valid_option_value_of_type_long(self):

            specifications =   (

                clasp.option('--length', alias='-l', value_type=long),
            )

            argv    =   ( 'myprog', '--length=123', '-l', '456', )
            args    =   clasp.parse(argv, specifications)

            self.assertEqual(0, len(args.flags))
            self.assertEqual(2, len(args.options))
            self.assertEqual(0, len(args.values))

            option  =   args.options[0]

            self.assertIsInstance(option, ( Option, ))
            self.assertEqual(option.given_index       ,   1)
            self.assertEqual(option.given_name        ,   '--length')
            self.assertTrue(option.argument_specification)
            self.assertEqual(option.given_hyphens     ,   2)
            self.assertEqual(option.given_label       ,   'length')
            self.assertEqual(option.name              ,   '--length')
            self.assertEqual(option.extras            ,   {})
            self.assertEqual(option.given_value       ,   '123')
            self.assertIsInstance(option.value        ,   long)
            self.assertEqual(option.value             ,   123)
            self.assertEqual(str(option)              ,   '--length=123')
            self.assertEqual(option                   ,   '--length=123')

            option  =   args.options[1]

            self.assertIsInstance(option, ( Option, ))
            self.assertEqual(option.given_index       ,   2)
            self.assertEqual(option.given_name        ,   '-l')
            self.assertTrue(option.argument_specification)
            self.assertEqual(option.given_hyphens     ,   1)
            self.assertEqual(option.given_label       ,   'l')
            self.assertEqual(option.name              ,   '--length')
            self.assertEqual(option.extras            ,   {})
            self.assertEqual(option.given_value       ,   '456')
            self.assertIsInstance(option.value        ,   long)
            self.assertEqual(option.value             ,   456)
            self.assertEqual(str(option)              ,   '--length=456')
            self.assertEqual(option                   ,   '--length=456')


    def test_invalid_option_value_of_type_int(self):

        specifications =   (

            clasp.option('--length', alias='-l', value_type=int),
        )

        argv    =   ( 'myprog', '--length=abc' )

        with self.assertRaises(clasp.InvalidIntegerException):

            clasp.parse(argv, specifications)


    def test_invalid_option_value_of_type_float(self):

        specifications =   (

            clasp.option('--length', alias='-l', value_type=float),
        )

        argv    =   ( 'myprog', '--length=abc' )

        with self.assertRaises(clasp.InvalidNumberException):

            clasp.parse(argv, specifications)


    def test_empty_option_value_of_type_int(self):

        specifications =   (

            clasp.option('--length', alias='-l', value_type=int),
        )

        argv    =   ( 'myprog', '--length=' )

        with self.assertRaises(clasp.MissingValueException):

            clasp.parse(argv, specifications)


    def test_empty_option_value_of_type_int_2(self):

        specifications =   (

            clasp.option('--length', alias='-l', value_type=int),
        )

        argv    =   ( 'myprog', '-l' )

        with self.assertRaises(clasp.MissingValueException):

            clasp.parse(argv, specifications)


    def test_empty_option_value_of_type_int_3(self):

        specifications =   (

            clasp.option('--length', alias='-l', value_type=int),
        )

        argv    =   ( 'myprog', '--length' )

        with self.assertRaises(clasp.MissingValueException):

            clasp.parse(argv, specifications)


    def test_empty_option_value_of_type_int_4(self):

        specifications =   (

            clasp.option('--length', alias='-l', value_type=int),
        )

        argv    =   ( 'myprog', '--length', '--' )

        with self.assertRaises(clasp.MissingValueException):

            clasp.parse(argv, specifications)


    def test_invalid_option_value_of_type_that_is_not_supported(self):

        with self.assertRaises(TypeError):

            clasp.option('--length', alias='-l', value_type='abc'),


    def test_flags_of_flags_and_options_combined(self):

        specifications =   (

            clasp.flag('--compile', alias='-c'),
            clasp.flag('--mode=debug', alias='-d'),
            clasp.flag('--execute', alias='-e'),
            clasp.option('--mode', alias='-m'),
            clasp.option('--length', alias='-l', value_type=int),
        )

        argv    =   ( 'myprog', '-ced', '-l', '123' )
        args    =   clasp.parse(argv, specifications)

        self.assertEqual('myprog', args.program_name)

        self.assertIsInstance(args.flags, ( tuple, ))
        self.assertTrue(args.flags)
        self.assertEqual(2, len(args.flags))

        flag    =   args.flags[0]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-ced')
        self.assertTrue(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'ced')
        self.assertEqual(flag.name              ,   '--compile')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '--compile')
        self.assertEqual(flag                   ,   '--compile')

        flag    =   args.flags[1]

        self.assertIsInstance(flag, ( Flag, ))
        self.assertEqual(flag.given_index       ,   1)
        self.assertEqual(flag.given_name        ,   '-ced')
        self.assertTrue(flag.argument_specification)
        self.assertEqual(flag.given_hyphens     ,   1)
        self.assertEqual(flag.given_label       ,   'ced')
        self.assertEqual(flag.name              ,   '--execute')
        self.assertEqual(flag.extras            ,   {})
        self.assertEqual(str(flag)              ,   '--execute')
        self.assertEqual(flag                   ,   '--execute')

        self.assertIsInstance(args.options, ( tuple, ))
        self.assertTrue(args.options)
        self.assertEqual(2, len(args.options))

        option  =   args.options[0]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index       ,   1)
        self.assertEqual(option.given_name        ,   '-ced')
        self.assertTrue(option.argument_specification)
        self.assertEqual(option.given_hyphens     ,   1)
        self.assertEqual(option.given_label       ,   'ced')
        self.assertEqual(option.name              ,   '--mode')
        self.assertEqual(option.extras            ,   {})
        self.assertEqual(option.given_value       ,   'debug')
        self.assertEqual(option.value             ,   'debug')
        self.assertEqual(str(option)              ,   '--mode=debug')
        self.assertEqual(option                   ,   '--mode=debug')

        option  =   args.options[1]

        self.assertIsInstance(option, ( Option, ))
        self.assertEqual(option.given_index       ,   2)
        self.assertEqual(option.given_name        ,   '-l')
        self.assertTrue(option.argument_specification)
        self.assertEqual(option.given_hyphens     ,   1)
        self.assertEqual(option.given_label       ,   'l')
        self.assertEqual(option.name              ,   '--length')
        self.assertEqual(option.extras            ,   {})
        self.assertEqual(option.given_value       ,   '123')
        self.assertEqual(option.value             ,   123)
        self.assertEqual(str(option)              ,   '--length=123')
        self.assertEqual(option                   ,   '--length=123')

        self.assertIsInstance(args.values, ( tuple, ))
        self.assertFalse(args.values)
        self.assertEqual(0, len(args.values))



if '__main__' == __name__:

    unittest.main()






from .exceptions import *
from .option_specification import OptionSpecification

from .util import _SUPPORT_long


_FALSE_STRINGS_lower = (

    "false",
    "no",
    "0",
)
_TRUE_STRINGS_lower = (

    "true",
    "yes",
    "1",
)


def _parse_to_bool(v):

    s   =   str(v)

    s   =   s.lower()

    if s in _FALSE_STRINGS_lower:

        return False

    if s in _TRUE_STRINGS_lower:

        return True

    return None


class OptionArgument(object):
    """
    TBC
    """


    def __init__(self, arg, given_index, given_name, resolved_name, argument_specification, given_hyphens, given_label, value, extras):

        if __debug__:
            if not argument_specification:

                pass
            elif isinstance(argument_specification, (OptionSpecification, )):

                pass
            else:

                raise TypeError("`argument_specification` must be instance of type `%s`; `%s` (%s) given" % (OptionSpecification, type(argument_specification), argument_specification))

        self.arg_           =   arg
        self.given_index    =   given_index
        self.given_name     =   given_name
        self.argument_specification =   argument_specification
        self.given_hyphens  =   given_hyphens
        self.given_label    =   given_label
        self.name           =   resolved_name if resolved_name else given_name
        self._set_value(value, True)
        self.extras         =   extras if extras else dict()

        self.private_fields =   {

            'used'  :   False
        }


    def _set_value(self, value, from_ctor=False):

        given_value         =   value

        arg_spec            =   self.argument_specification

        if arg_spec:

            if not given_value:

                given_value = arg_spec.default_value

            if given_value:


                if arg_spec.value_type:

                    assert arg_spec.value_type in OptionSpecification._VALID_VALUE_TYPES

                    if False:

                        pass
                    elif arg_spec.value_type == bool:

                        try:

                            value       =   _parse_to_bool(given_value)
                        except ValueError as x:

                            raise InvalidBooleanException("the '%s' option's value '%s' cannot be interpreted as boolean" % (self.name, given_value))
                    elif arg_spec.value_type == float:

                        try:

                            value       =   float(given_value)
                        except ValueError as x:

                            raise InvalidNumberException("the '%s' option's value '%s' cannot be interpreted as a number" % (self.name, given_value))
                    elif arg_spec.value_type == int:

                        try:

                            value       =   int(given_value)
                        except ValueError as x:

                            raise InvalidIntegerException("the '%s' option's value '%s' cannot be interpreted as an integer" % (self.name, given_value))
                    elif _SUPPORT_long and arg_spec.value_type == long:

                        try:

                            value       =   long(given_value)
                        except ValueError as x:

                            raise InvalidIntegerException("the '%s' option's value '%s' cannot be interpreted as a long integer" % (self.name, given_value))
                    elif arg_spec.value_type == str:

                        value   =   given_value
            else:

                if given_value is None and from_ctor:

                    pass
                else:

                    if False:

                        pass
                    elif arg_spec.value_type == str:

                        pass
                    elif not arg_spec.value_type:

                        pass
                    else:

                        raise MissingValueException("the '%s' option does not have a value to be interpreted as an integer" % (self.name))


        self.given_value    =   given_value
        self.value          =   value


    def use(self):

        self.private_fields['used'] = True


    def used(self):

        return self.private_fields['used']


    def __str__(self):

        if isinstance(self.value, (bool, )):

            v   =   str(self.value).lower()
        else:

            v   =   str(self.value)

        return "%s=%s" % (self.name, v)


    def __repr__(self):

        return "<%s.%s: given_index=%s; given_name=%s; given_value=%s; given_hyphens=%s, given_label=%s, extras=%s; argument_specification=%s >" %\
            (self.__module__, self.__class__.__name__, self.given_index, self.given_name, self.given_value, self.given_hyphens, self.given_label, self.extras, self.argument_specification, )


    def __eq__(self, other):
        """
        Yields `True` if `other` is a string that is the same as `#name`, or a `OptionArgument` or a `OptionSpecification` that has the same `#name`
        """

        if isinstance(other, OptionArgument):

            return str(self) == str(other)

        if isinstance(other, OptionSpecification):

            return self.name == other.name

        if isinstance(other, str):

            return str(self) == other

        return False


    def __ne__(self, other):
        """
        Yields `False` if `other` is not an `OptionArgument` or has a different `#name`
        """

        return not self.__eq__(other)






from .flag_specification import FlagSpecification

class FlagArgument(object):
    """
    TBC
    """


    def __init__(self, arg, given_index, given_name, resolved_name, argument_specification, given_hyphens, given_label, extras):

        if __debug__:
            if not argument_specification:

                pass
            elif isinstance(argument_specification, (FlagSpecification, )):

                pass
            else:

                raise TypeError("`argument_specification` must be instance of type `%s`; `%s` (%s) given" % (FlagSpecification, type(argument_specification), argument_specification))

        self.arg_           =   arg
        self.given_index    =   given_index
        self.given_name     =   given_name
        self.argument_specification =   argument_specification
        self.given_hyphens  =   given_hyphens
        self.given_label    =   given_label
        self.name           =   resolved_name if resolved_name else given_name
        self.extras         =   extras if extras else dict()

        self.private_fields =   {

            'used'  :   False
        }


    def use(self):

        self.private_fields['used'] = True


    def used(self):

        return self.private_fields['used']


    def __str__(self):

        return self.name


    def __eq__(self, other):
        """
        Yields `True` if `other` is a string that is the same as `#name`, or a `FlagArgument` or a `FlagSpecification` that has the same `#name`
        """

        if isinstance(other, FlagArgument):

            return self.name == other.name

        if isinstance(other, FlagSpecification):

            return self.name == other.name

        if isinstance(other, str):

            return self.name == other

        return False


    def __ne__(self, other):
        """
        Yields `False` if `other` is not a `FlagArgument` or has a different `#name`
        """

        return not self.__eq__(other)





#! /usr/bin/env python

# examples/duplicate_aliases.py


# imports

import pyclasp as clasp

import sys

# Specify specifications, parse, and checking standard flags

flag_Compile    =   clasp.flag('--compile', alias='-c')
flag_Debug      =   clasp.flag('--debug', alias='-d')
flag_Execute    =   clasp.flag('--execute', alias='-c')

specifications = (

    flag_Compile,
    flag_Debug,
    flag_Execute,

    clasp.HelpFlag(),
    clasp.VersionFlag(),
)

args = clasp.parse(sys.argv, specifications)


if args.flag_is_specified('--help'):

    clasp.show_usage(specifications, exit_code=0, stream=sys.stdout)
else:

    sys.stderr.write("specify the '--help' flag to observe the duplicate alias warning\n")


